/**
 * This class is the basic implementation of a color class to be used for TraX
 *
 * I have included the .mult, etc functions here because those seem
 *   to work faster without creating a new object on the stack to return
 *   from doing things like A * B (one can go nuts with templates and
 *   get that to be very optimized...)
 *
 * @author: Konstantin Shkurko (kshkurko@cs.utah.edu)
 * @date:	09.04.11
 */
#ifndef COLOR_H
#define COLOR_H

#include "trax.hpp"
#include "Constants.h"


/// basic color definition
class Color {

	public:
		/// basic constructor
		Color( const float &r1=0.f, const float &g1=0.f, const float &b1=0.f ) : r(r1), g(g1), b(b1) {}

		/// Copy constructor
		Color( const Color &clr ) {
			r = clr.r;		g = clr.g;		b = clr.b;
		}

		/// Constructor based on reading data from TRaX memory
		Color( const int &startMemAddress ) {
			set( startMemAddress );
		}

		/// basic setter
		inline void set( const Color &clr ) {
			r = clr.r;		g = clr.g;		b = clr.b;
		}

		/// basic setter
		inline void set( const float &r1=0.f, const float &g1=0.f, const float &b1=0.f ) {
			r = r1;			g = g1;			b = b1;
		}

		/// basic setter from TRaX memory location
		inline void set( const int &memAddr ) {
			r = loadf( memAddr, 0 );
			g = loadf( memAddr, 1 );
			b = loadf( memAddr, 2 );
		}

		/// Multiplication by another color
		inline void mult( const Color &clr ) {
			r *= clr.r;		g *= clr.g;		b *= clr.b;
		}

		/// Multiplication by a scalar
		inline void mult( const float &val=0.f ) {
			r *= val;		g *= val;		b *= val;
		}

		/// Addition of another color
		inline void add( const Color &clr ) {
			r += clr.r;		g += clr.g;		b += clr.b;
		}

		/// Subtraction of another color
		inline void sub( const Color &clr ) {
			r -= clr.r;		g -= clr.g;		b -= clr.b;
		}

		/// Clamp each color channel to [0,1]
		inline void clamp01() {
			r = max( min( r, 1.f ), 0.f );
			g = max( min( g, 1.f ), 0.f );
			b = max( min( b, 1.f ), 0.f );
		}

		/// Clamp each color channel to to [A,B]
		inline void clampAB( const float &A, const float &B ) {
			r = max( min( r, B ), A );
			g = max( min( g, B ), A );
			b = max( min( b, B ), A );
		}


		/// Assignment operator
		inline Color& operator=( const Color &clr ) {
			r = clr.r;		g = clr.g;		b = clr.b;
			return *this;
		}

		/// Multiplication by another color operator
		inline Color& operator*=( const Color &clr ) {
			r *= clr.r;		g *= clr.g;		b *= clr.b;
			return *this;
		}

		/// Multiplication by a scalar operator
		inline Color& operator*=( const float &val ) {
			r *= val;		g *= val;		b *= val;
			return *this;
		}

		/// Addition of another Color
		inline Color& operator+=( const Color &clr ) {
			r += clr.r;		g += clr.g;		b += clr.b;
			return *this;
		}

		/// Subtraction of another Color
		inline Color& operator-=( const Color &clr ) {
			r -= clr.r;		g -= clr.g;		b -= clr.b;
			return *this;
		}


	// public for speed
		float r;			/// red Color of this pixel
		float g;			/// green Color of this pixel
		float b;			/// blue Color of this pixel
};


// operators just for fun (performance-wise not so good)

/// Addition of two Colors
inline Color operator+( const Color &clr1, const Color &clr2 ) {
	return Color( clr1.r+clr2.r, clr1.g+clr2.g, clr1.b+clr2.b );
}

/// Subtraction of two Colors
inline Color operator-( const Color &clr1, const Color &clr2 ) {
	return Color( clr1.r-clr2.r, clr1.g-clr2.g, clr1.b-clr2.b );
}

/// Multiplication of two Colors
inline Color operator*( const Color &clr1, const Color &clr2 ) {
	return Color( clr1.r*clr2.r, clr1.g*clr2.g, clr1.b*clr2.b );
}

/// Multiplication by a scalar
inline Color operator*( const Color &clr, const float &val ) {
	return Color( clr.r*val, clr.g*val, clr.b*val );
}

/// Multiplication by a scalar
inline Color operator*( const float &val, const Color &clr ) {
	return Color( clr.r*val, clr.g*val, clr.b*val );
}

/// Clamps each Color channel to [0,1] range
inline Color clamp01( const Color &clr ) {
	return Color( MYRTC::clamp01( clr.r ),
				MYRTC::clamp01( clr.g ),
				MYRTC::clamp01( clr.b ) 
			);
}

/// Clamps each Color channel to [A,B] range
inline Color clampAB( const Color &clr, const float &A, const float &B ) {
	return Color( MYRTC::clampAB( clr.r, A, B ),
				MYRTC::clampAB( clr.g, A, B ),
				MYRTC::clampAB( clr.b, A, B ) 
			);
}

#endif // Color_H
