/**
 * Basic abstract class defining a lambertian material. This will be extended
 *   later to include specular and such materials.
 *
 * @author	Konstantin Shkurko (kshkurko@cs.utah.edu)
 * @date	09.10.11
 */
#ifndef MAT_PRIMITIVE_H
#define MAT_PRIMITIVE_H

#include "Color.h"


/// basic material primitive definition
class MatPrimitive {

	public:
		static const int SIZE_IN_TRAX = 25;			/// number of 4Bs (aka sizeof(float)) materials take up in TRaX memory

		/**
		 * Basic constructor
		 *
		 * @param diffClr - diffuse color
		 * @param kDiff	  - diffuse coefficient
		 * @param kAmb    - ambient coefficient
		 */
		MatPrimitive( const Color &diffClr=Color(1.f,1.f,1.f), const float &kDiff=0.7f, const float &kAmb=0.3f, 
						const float &matEta=1.0f, const bool &refl=false, const bool &refr=false ) :
							m_clrDiff(diffClr), m_fKdiff(kDiff), m_fKamb(kAmb), m_fMatEta(matEta),
							m_bFlagReflective(refl), m_bFlagRefractive(refr), m_bFlagProcTileSq(false)  {
		}


		/**
		 * Constructor from TRaX memory (just for diffuse color)
		 * @param addr  - address to load from 
		 * @param matID - id of the material to load
		 * @param kDiff - diffuse reflection coefficient
		 * @param kAmb  - ambient reflection coefficient
		 */
		MatPrimitive( const int &addr, const int &matID=0, const float &kDiff=0.7f, const float &kAmb=0.3f,
						const float &matEta=1.0f, const bool &refl=false, const bool &refr=false ) :
							m_fKdiff(kDiff), m_fKamb(kAmb), m_fMatEta(matEta), m_bFlagReflective(refl),
							m_bFlagRefractive(refr), m_bFlagProcTileSq(false) {
			set( addr, matID );
		}


		/**
		 * Basic setter
		 *
		 * @param diffClr - diffuse color
		 * @param kDiff	  - diffuse coefficient
		 * @param kAmb    - ambient coefficient
		 */
		inline void set( const Color &diffClr=Color(1.f,1.f,1.f), const float &kDiff=0.7f, const float &kAmb=0.3f,
							const float &matEta=1.0f, const bool &refl=false, const bool &refr=false) {
			m_clrDiff.set( diffClr );
			m_fKdiff  = kDiff;
			m_fKamb   = kAmb;
			m_fMatEta = matEta;
			m_bFlagReflective = refl;
			m_bFlagRefractive = refr;
		}


		/**
		 * Sets just the _diffuse_ color from TRaX memory
		 * @param addr  - address in memory for first material
		 * @param matID - id of the material to read in (adds to start address)
		 */
		inline void set( const int &addr, const int &matID=0 ) {
			const int startAddr = addr + SIZE_IN_TRAX*matID;
			m_clrDiff.set( startAddr + 4 );
		}


	// public for speed
		Color m_clrDiff;			/// diffuse color
		float m_fKdiff;				/// diffuse coefficient
		float m_fKamb;				/// ambient coefficient
		float m_fMatEta;			/// coefficient used for fresnel reflection and such
		bool m_bFlagReflective;		/// whether this object is reflective
		bool m_bFlagRefractive;		/// whether this object is refractive
		bool m_bFlagProcTileSq;		/// whether this object is procedural tile - square
};


#endif // MAT_PRIMITIVE_H
