/**
 * This class is the basic implementation of a vector class to be used for TraX
 *
 * I have included the .mult, etc functions here because those seem
 *   to work faster without creating a new object on the stack to return
 *   from doing things like A * B (one can go nuts with templates and
 *   get that to be very optimized...)
 *
 * @author: Konstantin Shkurko (kshkurko@cs.utah.edu)
 * @date:	09.04.11
 */
#ifndef VECTOR_H
#define VECTOR_H

#include "trax.hpp"
#include "Constants.h"


/// basic color definition
class Vector {

	public:
		/// basic constructor
		Vector( const float &x1=0.f, const float &y1=0.f, const float &z1=0.f ) : x(x1), y(y1), z(z1) {}

		/// Copy constructor
		Vector( const Vector &vec ) {
			x = vec.x;		y = vec.y;		z = vec.z;
		}

		// Constructor that loads vector from TRaX memory
		Vector( const int &address ) {
			set( address );
		}

		/// basic setter
		inline void set( const Vector &vec ) {
			x = vec.x;		y = vec.y;		z = vec.z;
		}

		/// basic setter
		inline void set( const float &x1=0.f, const float &y1=0.f, const float &z1=0.f ) {
			x = x1;			y = y1;			z = z1;
		}

		// load from TRaX memory
		inline void set( const int &address ) {
			x = loadf( address, 0 );
			y = loadf( address, 1 );
			z = loadf( address, 2 );
		}

		/// Multiplication by another color
		inline void mult( const Vector &vec ) {
			x *= vec.x;		y *= vec.y;		z *= vec.z;
		}

		/// Multiplication by a scalar
		inline void mult( const float &val=0.f ) {
			x *= val;		y *= val;		z *= val;
		}

		/// Addition of another color
		inline void add( const Vector &vec ) {
			x += vec.x;		y += vec.y;		z += vec.z;
		}

		/// Subtraction of another color
		inline void sub( const Vector &vec ) {
			x -= vec.x;		y -= vec.y;		z -= vec.z;
		}

		/// Normalize in place
		inline void normalize() {
			mult( invsqrt(lengthSq()) );
		}

		/// Get length squared
		inline float lengthSq() const {
			return( x*x + y*y + z*z );
		}

		/// Get length
		inline float length() const {
			return sqrt( lengthSq() );
		}


		/// Dot product
		inline float dot( const Vector &vec ) const {
			return( x*vec.x + y*vec.y + z*vec.z );
		}

		/// Cross product -> out = this x vec
		inline void cross( Vector &out, const Vector &vec ) const {
			out.x = y*vec.z - z*vec.y;
			out.y = z*vec.x - x*vec.z;
			out.z = x*vec.y - y*vec.x;
		}

		/// Clamp each color channel to [0,1]
		inline void clamp01() {
			x = max( min( x, 1.f ), 0.f );
			y = max( min( y, 1.f ), 0.f );
			z = max( min( z, 1.f ), 0.f );
		}

		/// Clamp each color channel to to [A,B]
		inline void clampAB( const float &A, const float &B ) {
			x = max( min( x, B ), A );
			y = max( min( y, B ), A );
			z = max( min( z, B ), A );
		}


		/// Assignment operator
		inline Vector& operator=( const Vector &vec ) {
			x = vec.x;		y = vec.y;		z = vec.z;
			return *this;
		}

		/// Multiplication by another color operator
		inline Vector& operator*=( const Vector &vec ) {
			x *= vec.x;		y *= vec.y;		z *= vec.z;
			return *this;
		}

		/// Multiplication by a scalar operator
		inline Vector& operator*=( const float &val ) {
			x *= val;		y *= val;		z *= val;
			return *this;
		}

		/// Addition of another Vector
		inline Vector& operator+=( const Vector &vec ) {
			x += vec.x;		y += vec.y;		z += vec.z;
			return *this;
		}

		/// Suztxaction of another Vector
		inline Vector& operator-=( const Vector &vec ) {
			x -= vec.x;		y -= vec.y;		z -= vec.z;
			return *this;
		}
		

	// public for speed
		float x;			/// x location of this Vector
		float y;			/// y location of this Vector
		float z;			/// z location of this Vector
};


// operators just for fun (performance-wise not so good)

/// Addition of two Vectors
inline Vector operator+( const Vector &vec1, const Vector &vec2 ) {
	return Vector( vec1.x+vec2.x, vec1.y+vec2.y, vec1.z+vec2.z );
}

/// Suztxaction of two Vectors
inline Vector operator-( const Vector &vec1, const Vector &vec2 ) {
	return Vector( vec1.x-vec2.x, vec1.y-vec2.y, vec1.z-vec2.z );
}

/// Multiplication of two Vectors
inline Vector operator*( const Vector &vec1, const Vector &vec2 ) {
	return Vector( vec1.x*vec2.x, vec1.y*vec2.y, vec1.z*vec2.z );
}

/// Multiplication by a scalar
inline Vector operator*( const Vector &vec, const float &val ) {
	return Vector( vec.x*val, vec.y*val, vec.z*val );
}

/// Multiplication by a scalar
inline Vector operator*( const float &val, const Vector &vec ) {
	return Vector( vec.x*val, vec.y*val, vec.z*val );
}

/// Clamps each Vector channel to [0,1] range
inline Vector clamp01( const Vector &vec ) {
	return Vector( MYRTC::clamp01( vec.x ),
				MYRTC::clamp01( vec.y ),
				MYRTC::clamp01( vec.z ) 
			);
}

/// Clamps each Vector channel to [A,B] range
inline Vector clampAB( const Vector &vec, const float &A, const float &B ) {
	return Vector( MYRTC::clampAB( vec.x, A, B ),
				MYRTC::clampAB( vec.y, A, B ),
				MYRTC::clampAB( vec.z, A, B ) 
			);
}

#endif // Vector_H
