/**
 * Implementation of the basic sphere class.
 *
 * A sphere is defined as center and radius
 *
 * @author:	Konstantin Shkurko (kshkruko@cs.utah.edu)
 * @date:	09.04.11
 */
#ifndef GP_SPHERE_H
#define GP_SPHERE_H

#include "Color.h"                      /// for now we'll stick the color in here to be replaced by material
#include "Vector.h"
#include "RayPrimitive.h"
#include "MatPrimitive.h"

// dummy class -> "IntersectionRecord.h" must be included in every derived geometry class
class IntersectionRecord;


/// define basic sphere
class gpSphere {

	public:
		/**
		 * Basic constructor which creates a sphere at (0,0,0) with radius 1
		 * @param ctr - center of this sphere
		 * @param rad - radius of this sphere
		 */
		gpSphere( const Vector &ctr=Vector(0.f,0.f,0.f), const float &rad=1.0f) : 
							m_vCenter(ctr), m_pMat(0) {
			m_fRadInv = 1.f/rad;
			m_fRadSq  = rad*rad;
		}


		/**
		 * Basic constructor
		 * @param cx  - x coordinate of sphere center
		 * @param cy  - y coordinate of sphere center
		 * @param cz  - z coordinate of sphere center
		 * @param rad - new radius of the sphere
		 */
		gpSphere( const float &cx, const float &cy, const float &cz, const float &rad ) : m_pMat(0) {
			set( cx, cy, cz, rad );
		}


		/**
		 * Sets the information for this sphere
		 * @param ctr - new center of the sphere
		 * @param rad - new radius of the sphere
		 */
		inline void set( const Vector &ctr=Vector(0.f,0.f,0.f), const float &rad=1.0f ) {
			m_vCenter.set( ctr );
			m_fRadInv = 1.f/rad;
			m_fRadSq  = rad*rad;
		}


		/**
		 * Sets the information for this sphere
		 * @param cx  - x coordinate of sphere center
		 * @param cy  - y coordinate of sphere center
		 * @param cz  - z coordinate of sphere center
		 * @param rad - new radius of the sphere
		 */
		inline void set( const float &cx, const float &cy, const float &cz, const float &rad ) {
			m_vCenter.set( cx, cy, cz );
			m_fRadInv = 1.f/rad;
			m_fRadSq  = rad*rad;
		}


		/**
		 * Get normal based on the hit location
		 * 
		 * @param hitNorm  - normal of the hit on sphere
		 * @param hitLoc   - location of the hit on sphere
		 */
		void getNormal( Vector &hitNorm, const Vector &hitLoc ) const;


		/**
		 * Intersection between a ray and this primitive - this one provides
		 *   detailed intersection information (use for primary rays!)
		 *
		 * @param ray - ray to intersect against (MUST BE NORMALIZED)
		 * @param rec - intersection record holding intersection information
		 * @return flag if we actually intersected
		 */
		bool intersectDetailed( RayPrimitive &ray, IntersectionRecord &rec ) const;


		/**
		 * Intersection between a ray and this primitive - this one provides
		 *   quick intersection test (use for shadow rays!)
		 *
		 * @param ray - ray to intersect against (not normalized)
		 * @param t0  - computed intersection 1
		 * @param t1  - computed intersection 2
		 * @return flag if we actually intersected
		 */
		bool intersectQuick( const RayPrimitive &ray ) const;


		/**
		 * Intersection between a ray and this primitive - this one provides
		 *   quick intersection test (use for shadow rays!)
		 *
		 * @param ray - ray to intersect against (normalized)
		 * @param t0  - computed intersection 1
		 * @param t1  - computed intersection 2
		 * @return flag if we actually NOT intersected
		 */
		bool notIntersectQuickNormed( const RayPrimitive &ray ) const;


	// public for speed
		Vector m_vCenter;			/// center of this sphere
		float  m_fRadSq;			/// square of the radius (to accelerate intersections)
		float  m_fRadInv;			/// inverse radius of this sphere
		MatPrimitive* m_pMat;		/// pointer to the material
};


#endif // GP_SPHERE_H
