/**
 * Implementation of the basic triangle class - just one triangle
 *   This should be used for now, since we'll have a few of them, but
 *   soon we'll need to create a triangle mesh class to save on storage
 *
 * @author	Konstantin Shkurko (kshkurko@cs.utah.edu)
 * @date	09.12.11
 */
#ifndef GP_TRIANGLE_H
#define GP_TRIANGLE_H

#include "trax.hpp"
#include "Color.h"                      /// for now we'll stick the color in here to be replaced by material
#include "Vector.h"
#include "RayPrimitive.h"
#include "MatPrimitive.h"

// dummy class -> "IntersectionRecord.h" must be included in every derived geometry class
class IntersectionRecord;


// triangle class
class gpTriangle {

	public:
		static const int SIZE_IN_TRAX = 11;				/// number of 4Bs (aka sizeof(float)) this triangle takes up

		/**
		 * Basic constructor. Triangle normal is defined in the usual manner
		 *   based on the vertex order
		 *
		 * @param p0 - vertex 0
		 * @param p1 - vertex 1
		 * @param p2 - vertex 2
		 */
		gpTriangle( const Vector &p0=Vector(1.f,0.f,0.f), const Vector &p1=Vector(0.f,1.f,0.f),
					const Vector &p2=Vector(0.f,0.f,1.f) ) : m_objAddr(0), m_matID(0) {

			set( p0, p1, p2 );
#ifdef DO_MULTIOBJECT
			m_pMat = 0;
#endif
		}


		/**
		 * Sets the triangle locations and such
		 *
		 * @param p0 - vertex 0
		 * @param p1 - vertex 1
		 * @param p2 - vertex 2
		 */
		inline void set( const Vector &p0=Vector(1.f,0.f,0.f), const Vector &p1=Vector(0.f,1.f,0.f),
						const Vector &p2=Vector(0.f,0.f,1.f) ) {
			m_vPoint0.set( p0 );
			m_vPoint1.set( p1 );
			m_vPoint2.set( p2 );
			// do we need to adjust from points to edges? (this makes starting triangles human-readable)
#if TRIANGLES_STORE_EDGES==1
			m_vPoint0.sub( p1 );
			m_vPoint2.sub( p1 );
#endif
		}


		/**
		 * Sets this triangle from memory
		 * @param addr - start address of this triangle
		 */
		inline void set( const int &addr ) {
			m_vPoint0.set( addr );
			m_vPoint1.set( addr + 3 );
			m_vPoint2.set( addr + 6 );

			m_objAddr = addr;
			m_matID   = loadi( addr, 10 );
		}


		/**
		 * Sets this triangle from memory BUT only
		 *   the vertecies to save on loading extraneous info
		 * This should be used for shadow-ray intersections
		 * @param addr - start address of this triangle
		 */
		inline void setQuick( const int &addr ) {
			m_objAddr = addr;
			m_vPoint0.set( addr );
			m_vPoint1.set( addr + 3 );
			m_vPoint2.set( addr + 6 );
		}


		/**
		 * Intersection between a ray and a this triangle - this one provides
		 *   detailed intersection information (use for primary rays!)
		 *
		 * @param ray - ray to intersect against
		 * @param rec - intersection record holding intersection information
		 * @return flag if we actually intersected
		 */
		bool intersectDetailed( RayPrimitive &ray, IntersectionRecord &rec ) const;


		/**
		 * Intersection between a ray and this triangle - this one provides
		 *   quick intersection test (use for shadow rays!)
		 *
		 * @param ray - ray to intersect against
		 * @return flag if we actually intersected
		 */
		bool intersectQuick( const RayPrimitive &ray ) const;


		/**
		 * Get normal based on the hit location
		 * 
		 * @param hitNorm  - normal of the hit on sphere
		 * @param hitLoc   - location of the hit on sphere
		 */
		inline void getNormal( Vector &hitNorm, const Vector &hitLoc ) const {
#if TRIANGLES_STORE_EDGES==1
			// treat data p0 and p2 as edges to compute the normal
			m_vPoint2.cross( hitNorm, m_vPoint0 );		// normal = edge2 x edge1
#else
			// treat data p0 and p2 as triangle vertices
			Vector edge1( m_vPoint0 );
			edge1.sub( m_vPoint1 );
			Vector edge2( m_vPoint2 );
			edge2.sub( m_vPoint1 );
			edge2.cross( hitNorm, edge1 );				// normal = edge2 x edge1
#endif
			hitNorm.normalize();
		}


	// public for speed
		Vector m_vPoint1;			/// central point, p1
		Vector m_vPoint0;			/// left point p0, or p0-p1 edge (not-normalized)
		Vector m_vPoint2;			/// right point p2, or p2-p1 edge (not-normalized)

		// now this is some TRaX stuff for loading scenes -> perhaps make a separate triangle class?
		int m_objAddr;				/// TRaX triangle address in memory (saves + and *)
		int m_matID;				/// TRaX material id (not an address!)

#ifdef DO_MULTIOBJECT
		MatPrimitive* m_pMat;		/// pointer to the material
#endif
};


#endif // GP_TRI_H
