/**
 * Basic class describing a camera (eye location, target, fov etc)
 * 
 *
 * @author	Konstantin Shkurko (kshkurko@cs.utah.edu)
 * @date	09.05.11
 */

#ifndef MY_CAMERA_H
#define MY_CAMERA_H

#include "Constants.h"
#include "Vector.h"
#include "RayPrimitive.h"


/// basic camera class
class myCamera {

	public:
		/**
		 * Constructor
		 *
		 * @param eye    - location of camera in world space
		 * @param target - the point in space camera is looking at
		 * @param up     - suggested up direction for camera
		 * @param fov    - horizontal field of view of the camera in deg (actually: tan(fov_rad/2))
		 */
		myCamera( const Vector &eye=Vector(0.f,0.f,-5.f), const Vector &target=Vector(0.f,0.f,0.f), 
				const Vector &up=Vector(0.f,1.f,0.f), const float &width=512.f,
						const float &height=512.f, const float &fov_uLen=MYRTC::CAM_FOV_60_DEG ) :
						m_fvEye(eye), m_fvTarget(target), m_fFov_ULen(fov_uLen) {
		
			m_fAspectRatInv = width / height;

			// construct our basis
			constructFrame( up );
		}


		/**
		 * Constructor for the camera, which reads the information from the
		 *   TRaX scene memory.
		 *
		 * order in memory is: center, (lower,left image) corner, across (scaled right),
		 *     up (scaled up), unitgaze (front), u (right), v (up), radius (scalar)
		 * NOTE: see SimHWRT/trunk/Camera.cc
		 * 
		 * @param startMemAddress - start memory address for the camera
		 */
		myCamera( const int &startMemAddress ) {
			m_fvEye.set( startMemAddress );			// eye
			m_fvTarget.set( startMemAddress+3 );	// corner (bottom-left pt of image) -> incodes target
			m_fvRight.set( startMemAddress+6 );		// right = across
			m_fvUp.set( startMemAddress+9 );		// up = up
			m_fvFront.set( startMemAddress+12 );	// gaze = front direction
			m_fvRightImg.set( startMemAddress+15 );	// img right = u
			m_fvUpImg.set( startMemAddress+18 );	// img up = v

			// get the aspect ratio from lengths of across and up :D
			const float w    = m_fvRight.length();
			const float hInv = 1.f / m_fvUp.length();
			m_fAspectRatInv  = w*hInv;

			// get the fov_ULen -> which is w? from above
			m_fFov_ULen = w;

			// normalize right and up (they hold the total image encoding)
			m_fvRight.mult( 1.f/w );
			m_fvUp.mult( hInv );

			// compute target location from corner
			Vector tmp( m_fvRightImg );
			tmp.add( m_fvUpImg );
			tmp.mult( 0.5f );
			m_fvTarget.add( tmp );
		}


		/**
		 * Sets the entire camera just like the constructor
		 *
		 * @param eye    - location of camera in world space
		 * @param target - the point in space camera is looking at
		 * @param up     - suggested up direction for camera
		 * @param fov    - horizontal field of view of the camera in deg (actually: tan(fov_rad/2))
		 */
		inline void set( const Vector &eye=Vector(0.f,0.f,-5.f), const Vector &target=Vector(0.f,0.f,0.f), 
						const Vector &up=Vector(0.f,1.f,0.f), const float &width=512.f,
						const float &height=512.f, const float &fov_uLen=MYRTC::CAM_FOV_60_DEG ) {
					
			m_fvEye.set( eye );
			m_fvTarget.set( target );
			m_fFov_ULen		= fov_uLen;
			m_fAspectRatInv	= height / width;
	
			constructFrame( up );
		}


		/**
		 * Constructs the correct orthonormal (left handed) coordinate frame for this camera
		 *   aka computes the up, right and back vectors
		 *
		 * @param up - suggested up vector for the camera
		 */
		inline void constructFrame( const Vector &up=Vector(0.f,1.f,0.f) ) {
			
			m_fvFront = m_fvTarget - m_fvEye;
			m_fvFront.normalize();

			// check whether our suggested vector is poor = almost parallel to front
			m_fvUp.set( up );
			m_fvUp.normalize();
			if( m_fvFront.dot( m_fvUp ) > 0.99999f ) {
				m_fvUp += Vector( 0.1f, 0.1f, 0.1f );
				m_fvUp.normalize();
			}

			// compute the left-handed coordinate system
			m_fvFront.cross( m_fvRight, up );		// R = F x up
			m_fvRight.normalize();
			m_fvRight.cross( m_fvUp, m_fvFront );	// U = R x F
			//m_fvUp.normalize();

			// scale the image up/right vectors
			m_fvRightImg.set( m_fvRight );
			m_fvRightImg.mult( m_fFov_ULen );
			m_fvUpImg.set( m_fvUp );
			m_fvUpImg.mult( (m_fFov_ULen*m_fAspectRatInv) );
		}


		/**
		 * Sets the new image size and new FOV based on the input
		 */
		inline void setImgSizeFOV( const float &w=512.f, const float &h=512.f, 
									const float &fov_ulen=MYRTC::CAM_FOV_60_DEG ) {

			// recompute new aspect ratio and reset fov
			m_fFov_ULen		= fov_ulen;
			m_fAspectRatInv	= h/w;

			// update our right/up images
			m_fvRightImg.set( m_fvRight );
			m_fvRightImg.mult( m_fFov_ULen );
			m_fvUpImg.set( m_fvUp );
			m_fvUpImg.mult( (m_fFov_ULen*m_fAspectRatInv) );
		}


		/**
		 * Generates a new ray and returns it. Ray's origin is camera's eye
		 *   ray's direction is: front + x*right + y*up
		 *
		 * right and up have been rescaled by FOV and aspect ratio. x,y are in [-1,1]
		 *
		 * @param x   - x direction of ray
		 * @param y   - y direction of ray
		 * @param ray - output ray
		 */
		virtual inline void generateRay( RayPrimitive &ray, const float &x=0.f, const float &y=0.f ) const {

			ray.m_vOrigin.set( m_fvEye );
			generateRayDirection( ray, x, y );
		}


		/**
		 * Updates a ray direction based on x,y directions
		 *   ray's direction is: front + x*right + y*up
		 *
		 * Both x,y are in [-1,1]
		 *
		 * @param x   - x direction of ray
		 * @param y   - y direction of ray
		 * @param ray - output ray
		 */
		virtual inline void generateRayDirection( RayPrimitive &ray, const float &x=0.f, const float &y=0.f ) const {

			Vector rit( m_fvUpImg );
			rit.mult( y );
			ray.m_vDirection.set( m_fvRightImg );
			ray.m_vDirection.mult( x );
			ray.m_vDirection.add( rit );
			ray.m_vDirection.add( m_fvFront );
			ray.m_vDirection.normalize();
		}


		/**
		 * Gets the eye location of this camera
		 */
		inline const Vector& getEye() const {
			return m_fvEye;
		}

		/**
		 * Gets the target location of this camera
		 */
		inline const Vector& getTarget() const {
			return m_fvTarget;
		}

		/**
		 * Gets the up vector for this camera
		 */
		inline const Vector& getUp() const {
			return m_fvUp;
		}

		/**
		 * Gets the right vector for this camera
		 */
		inline const Vector& getRight() const {
			return m_fvRight;
		}

		/**
		 * Gets the front vector for this camera
		 */
		inline const Vector& getFront() const {
			return m_fvFront;
		}

		/**
		 * Gets the field of view for this camera
		 */
		inline const float& getFOV_ULen() const {
			return m_fFov_ULen;
		}

	public:
		Vector m_fvFront;		/// front vector (normalized, this is a left-handed coordinate system)
		Vector m_fvRightImg;	/// scaled right vefore for ray generation
		Vector m_fvUpImg;		/// scaled up vector for ray generation
		Vector m_fvEye;			/// eye
		Vector m_fvTarget;		/// target for camera
		Vector m_fvUp;			/// up vector (normalized)
		Vector m_fvRight;		/// right vector (normalized)
		float m_fFov_ULen;		/// horizontal field of view in terms of length = tan(fovDeg/2*pi/180)
		float m_fAspectRatInv;	/// output image aspect ratio (height/width)
};

#endif // DEF_BOD_CAMERA_H
